import unittest
import numpy.testing as np_tst
from .....vassal.analysis.PairDistanceAnalysis import PairDistanceAnalysis
from .....vassal.analysis.voronoi.VoronoiFace import VoronoiFace
from .....vassal.data.Atom import Atom
from .....vassal.data.AtomImage import AtomImage
from .....vassal.data.Cell import Cell
import math

class testVoronoiFace(unittest.TestCase):
    def setUp(self):
        # Make a simple crystal.
        self.cell = Cell()
        self.cell.add_atom(Atom([0, 0, 0], 0))

    def tearDown(self):
        self.cell = None

    def test_initialize(self):
        # Initialize face.
        image = AtomImage(self.cell.get_atom(0), [0, 0, 1])
        face = VoronoiFace(self.cell.get_atom(0), image, radical=True)

        # Check out properties.
        np_tst.assert_array_almost_equal([0, 0, 0.5], face.get_face_center())
        self.assertAlmostEquals(0.5, face.get_face_distance(), delta=1e-6)
        np_tst.assert_array_almost_equal([0, 0, 1], face.get_normal())

    def test_simple_assemble(self):
        # Create face to be assembled.
        image = AtomImage(self.cell.get_atom(0), [0, 0, 1])
        face_to_assemble = VoronoiFace(self.cell.get_atom(0), image,
                                      radical=True)

        # Initialize faces.
        faces = []
        faces.append(VoronoiFace(self.cell.get_atom(0), AtomImage(
            self.cell.get_atom(0), [-1, 0, 0]), radical=True))
        faces.append(VoronoiFace(self.cell.get_atom(0), AtomImage(
            self.cell.get_atom(0), [1, 0, 0]), radical=True))
        faces.append(VoronoiFace(self.cell.get_atom(0), AtomImage(
            self.cell.get_atom(0), [0, -1, 0]), radical=True))
        faces.append(VoronoiFace(self.cell.get_atom(0), AtomImage(
            self.cell.get_atom(0), [0, 1, 0]), radical=True))


        # Attempt to assemble face.
        self.assertTrue(face_to_assemble.assemble_face_from_faces(faces))

        # Check out properties.
        self.assertEquals(4, face_to_assemble.n_edges())
        self.assertAlmostEquals(1.0, face_to_assemble.get_area(), delta=1e-6)

    def test_FCC_simple_assemble(self):
        # Make FCC crystal.
        self.cell.add_atom(Atom([0.5, 0.5, 0.0], 0))
        self.cell.add_atom(Atom([0.5, 0.0, 0.5], 0))
        self.cell.add_atom(Atom([0.0, 0.5, 0.5], 0))

        # Create face to be assembled.
        image = AtomImage(self.cell.get_atom(1), [0, 0, 0])
        face_to_assemble = VoronoiFace(self.cell.get_atom(0), image,
                                       radical=True)

        # Initialize faces that will be its neighbors.
        neighbor_faces = []
        neighbor_faces.append(VoronoiFace(self.cell.get_atom(0), AtomImage(
            self.cell.get_atom(2), [0, 0, 0]), radical=True))
        neighbor_faces.append(VoronoiFace(self.cell.get_atom(0), AtomImage(
            self.cell.get_atom(2), [0, 0, -1]), radical=True))
        neighbor_faces.append(VoronoiFace(self.cell.get_atom(0), AtomImage(
            self.cell.get_atom(3), [0, 0, 0]), radical=True))
        neighbor_faces.append(VoronoiFace(self.cell.get_atom(0), AtomImage(
            self.cell.get_atom(3), [0, 0, -1]), radical=True))

        # Attempt to assemble face.
        self.assertTrue(face_to_assemble.assemble_face_from_faces(
            neighbor_faces))

        # Checkout properties.
        self.assertEquals(4, face_to_assemble.n_edges())

        # Now attempt to assemble face considering all possible neighbors.
        neighbor_faces.append(VoronoiFace(self.cell.get_atom(0), AtomImage(
            self.cell.get_atom(1), [0, 0, 0]), radical=True))
        neighbor_faces.append(VoronoiFace(self.cell.get_atom(0), AtomImage(
            self.cell.get_atom(1), [0, -1, 0]), radical=True))
        neighbor_faces.append(VoronoiFace(self.cell.get_atom(0), AtomImage(
            self.cell.get_atom(1), [-1, 0, 0]), radical=True))
        neighbor_faces.append(VoronoiFace(self.cell.get_atom(0), AtomImage(
            self.cell.get_atom(1), [-1, -1, 0]), radical=True))
        neighbor_faces.append(VoronoiFace(self.cell.get_atom(0), AtomImage(
            self.cell.get_atom(2), [-1, 0, 0]), radical=True))
        neighbor_faces.append(VoronoiFace(self.cell.get_atom(0), AtomImage(
            self.cell.get_atom(2), [-1, 0, -1]), radical=True))
        neighbor_faces.append(VoronoiFace(self.cell.get_atom(0), AtomImage(
            self.cell.get_atom(3), [0, -1, 0]), radical=True))
        neighbor_faces.append(VoronoiFace(self.cell.get_atom(0), AtomImage(
            self.cell.get_atom(3), [0, -1, -1]), radical=True))

        # Attempt to assemble face.
        self.assertTrue(face_to_assemble.assemble_face_from_faces(
            neighbor_faces))

        # Checkout properties.
        self.assertEquals(4, face_to_assemble.n_edges())

    def test_FCC_direct_assemble(self):
        # Make FCC crystal.
        self.cell.add_atom(Atom([0.5, 0.5, 0.0], 0))
        self.cell.add_atom(Atom([0.5, 0.0, 0.5], 0))
        self.cell.add_atom(Atom([0.0, 0.5, 0.5], 0))

        # Create face to be assembled.
        image = AtomImage(self.cell.get_atom(1), [0, 0, 0])
        face_to_assemble = VoronoiFace(self.cell.get_atom(0), image,
                                       radical=True)

        # Initialize faces that are direct neighbors of atom 0.
        neighbor_faces = []
        neighbor_faces.append(VoronoiFace(self.cell.get_atom(0), AtomImage(
            self.cell.get_atom(1), [0, 0, 0]), radical=True))
        neighbor_faces.append(VoronoiFace(self.cell.get_atom(0), AtomImage(
            self.cell.get_atom(1), [0, -1, 0]), radical=True))
        neighbor_faces.append(VoronoiFace(self.cell.get_atom(0), AtomImage(
            self.cell.get_atom(1), [-1, 0, 0]), radical=True))
        neighbor_faces.append(VoronoiFace(self.cell.get_atom(0), AtomImage(
            self.cell.get_atom(1), [-1, -1, 0]), radical=True))
        neighbor_faces.append(VoronoiFace(self.cell.get_atom(0), AtomImage(
            self.cell.get_atom(2), [-1, 0, 0]), radical=True))
        neighbor_faces.append(VoronoiFace(self.cell.get_atom(0), AtomImage(
            self.cell.get_atom(2), [-1, 0, -1]), radical=True))
        neighbor_faces.append(VoronoiFace(self.cell.get_atom(0), AtomImage(
            self.cell.get_atom(2), [0, 0, 0]), radical=True))
        neighbor_faces.append(VoronoiFace(self.cell.get_atom(0), AtomImage(
            self.cell.get_atom(2), [0, 0, -1]), radical=True))
        neighbor_faces.append(VoronoiFace(self.cell.get_atom(0), AtomImage(
            self.cell.get_atom(3), [0, 0, 0]), radical=True))
        neighbor_faces.append(VoronoiFace(self.cell.get_atom(0), AtomImage(
            self.cell.get_atom(3), [0, 0, -1]), radical=True))
        neighbor_faces.append(VoronoiFace(self.cell.get_atom(0), AtomImage(
            self.cell.get_atom(3), [0, -1, 0]), radical=True))
        neighbor_faces.append(VoronoiFace(self.cell.get_atom(0), AtomImage(
            self.cell.get_atom(3), [0, -1, -1]), radical=True))

        # Attempt to assemble face.
        self.assertTrue(face_to_assemble.assemble_face_from_faces(
            neighbor_faces))

        # Checkout properties.
        self.assertEquals(4, face_to_assemble.n_edges())

        # Checkout properties.
        self.assertEquals(4, face_to_assemble.n_edges())

        # Now, see if all of the faces assemble.
        volume = 0.0
        for i, face in enumerate(neighbor_faces):
            self.assertTrue(face.assemble_face_from_faces(neighbor_faces))
            self.assertEquals(4, face.n_edges())
            # print i, face.get_outside_atom(), face.get_area(), \
            #     face.get_face_distance()
            area = face.get_area()
            volume += face.get_area() * face.get_face_distance() / 3.0

        self.assertAlmostEquals(0.25, volume, delta=1e-6)

    def test_FCC_full_assemble(self):
        # Make FCC crystal.
        self.cell.add_atom(Atom([0.5, 0.5, 0.0], 0))
        self.cell.add_atom(Atom([0.5, 0.0, 0.5], 0))
        self.cell.add_atom(Atom([0.0, 0.5, 0.5], 0))

        # Create face to be assembled.
        image = AtomImage(self.cell.get_atom(1), [0, 0, 0])
        face_to_assemble = VoronoiFace(self.cell.get_atom(0), image,
                                       radical=True)

        # Get neighbor finding tool.
        imageFinder = PairDistanceAnalysis()
        imageFinder.set_cutoff_distance(3.0)
        imageFinder.analyze_structure(self.cell)

        # Initialize any face.
        neighbor_faces = []
        for new_image in imageFinder.get_all_neighbors_of_atom(0):
            face = VoronoiFace(self.cell.get_atom(0), new_image[0],
                               radical=True)
            neighbor_faces.append(face)

        # Attempt to assemble face.
        self.assertTrue(face_to_assemble.assemble_face_from_faces(
            neighbor_faces))

        # Checkout properties.
        self.assertEquals(4, face_to_assemble.n_edges())

    def test_intersections(self):
        # Create test face.
        image = AtomImage(self.cell.get_atom(0), [0, 0, 2])
        test_face = VoronoiFace(self.cell.get_atom(0), image,
                                       radical=True)

        # Initialize faces.
        original_faces = []
        original_faces.append(VoronoiFace(self.cell.get_atom(0), AtomImage(
            self.cell.get_atom(0), [-1, 0, 0]), radical=True))
        original_faces.append(VoronoiFace(self.cell.get_atom(0), AtomImage(
            self.cell.get_atom(0), [2, 0, 0]), radical=True))
        original_faces.append(VoronoiFace(self.cell.get_atom(0), AtomImage(
            self.cell.get_atom(0), [0, -1, 0]), radical=True))
        original_faces.append(VoronoiFace(self.cell.get_atom(0), AtomImage(
            self.cell.get_atom(0), [0, 1, 0]), radical=True))
        test_face.assemble_face_from_faces(original_faces)

        self.assertAlmostEquals(1.5, test_face.get_area(), delta=1e-6)

        # Ensure that a non-intersecting face returns "None".
        self.assertEquals(None, test_face.compute_intersection(VoronoiFace(
            self.cell.get_atom(0), AtomImage(self.cell.get_atom(0), [3, 0,
                                                        0]), radical=True)))

        # Check intersection with a face that cuts off an edge completely.
        new_face = VoronoiFace(self.cell.get_atom(0), AtomImage(
            self.cell.get_atom(0), [1, 0, 0]), radical=True)
        self.assertAlmostEquals(1.0, test_face.get_cut_length(new_face),
                                delta=1e-6)

        new_edge = test_face.compute_intersection(new_face)
        self.assertTrue(new_edge is not None)
        self.assertEquals(new_face, new_edge.get_edge_face())
        self.assertEquals(test_face, new_edge.get_intersecting_face())
        self.assertTrue(test_face.is_closed())
        self.assertAlmostEquals(1.0, test_face.get_area(), delta=1e-6)

        # Compute intersection with a face that adds a new edge.
        test_face.assemble_face_from_faces(original_faces)
        new_face = VoronoiFace(self.cell.get_atom(0), AtomImage(
            self.cell.get_atom(0), [1, 1, 0]), radical=True)
        self.assertAlmostEquals(math.sqrt(2)/2, test_face.get_cut_length(
            new_face), delta=1e-6)

        new_edge = test_face.compute_intersection(new_face)
        self.assertTrue(new_edge is not None)
        self.assertEquals(new_face, new_edge.get_edge_face())
        self.assertEquals(test_face, new_edge.get_intersecting_face())
        self.assertEquals(5, test_face.n_edges())


import numpy as np

class ConstantColumns(object):
    """
    remove constant columns


    Attributes
    ----------
    removed_columns: list of column headers that have been removed

    Returns
    -------
    df: pandas dataframe

    """
    def fit_transform(self,df):
        """
        fit the method to the input dataframe and change it accordingly

        Parameters
        ----------
        df: pandas dataframe
            input dataframe

        Returns
        -------
        transformed dataframe
        """
        dfc = df.columns
        df = df.loc[:, (df != df.ix[0]).any()]
        self.removed_columns_ = np.array([i for i in dfc if i not in df.columns])
        return df
    def transform(self,df):
        """
        find and remove headers that are in the removed_columns_ attribute of the previous fit_transform method

        Parameters
        ----------
        df: pandas dataframe
            input dataframe

        Returns
        -------
        transformed dataframe
        """
        df = df.drop(self.removed_columns_,1)
        return df

class Outliers(object):
    """
    remove all rows where the values of a certain column are within an specified
    standard deviation from mean/median.

    Parameters
    ----------
    m: float, optional (default=3.0)
        the outlier threshold with respect to the standard deviation

    strategy: string, optional (default='median')
        available options: 'mean' and 'median'
        Values of each column will be compared to the 'mean' or 'median' of that column.

    Attributes
    ----------
    removed_rows_: numpy array of indices that have been removed

    Notes
    -----
    We highly recommend you to remove constant columns first and then remove outliers.
    """
    def __init__(self, m=2.0, strategy = 'median'):
        self.m = m
        self.strategy = strategy
    def fit_transform(self,df):
        """
        fit the method to the input dataframe and change it accordingly

        Parameters
        ----------
        df: pandas dataframe
            input dataframe

        Returns
        -------
        transformed dataframe
        """
        if self.strategy == 'mean':
            mask = ((df - df.mean()).abs() <= self.m * df.std(ddof=0)).T.all()
        elif self.strategy == 'median':
            mask = (((df - df.median()).abs()) <= self.m * df.std(ddof=0)).T.all()
        df = df.loc[mask, :]
        self.removed_rows_ = np.array(mask[mask==False].index)
        return df
    def transform(self,df):
        """
        find and remove rows/indices that are in the removed_rows_ attribute of the previous fit_transform method

        Parameters
        ----------
        df: pandas dataframe
            input dataframe

        Returns
        -------
        transformed dataframe
        """
        df = df.drop(self.removed_rows_,0)
        return df



import unittest
from ....data.materials.CompositionEntry import CompositionEntry
from ....utility.tools.IonicCompoundFinder import IonicCompoundFinder

class testIonicCompoundFinder(unittest.TestCase):
    def setUp(self):
        self.icf = IonicCompoundFinder()

    def tearDown(self):
        self.icf = None

    def test_NaCl(self):
        self.icf.set_nominal_composition(CompositionEntry(composition="NaCl"))
        self.icf.set_maximum_distance(0.2)
        self.icf.set_max_formula_unit_size(4)

        # Make sure it finds only one.
        accepted = self.icf.find_all_compounds()
        self.assertEquals(1, len(accepted))

    def test_FeO(self):
        self.icf.set_nominal_composition(CompositionEntry(composition="Fe2O"))
        self.icf.set_maximum_distance(0.34)
        self.icf.set_max_formula_unit_size(5)

        # Make sure it finds only one (FeO).
        accepted = self.icf.find_all_compounds()
        self.assertEquals(1, len(accepted))

        # Make sure it finds two (FeO, Fe2O3).
        self.icf.set_maximum_distance(0.54)
        accepted = self.icf.find_all_compounds()
        self.assertEquals(2, len(accepted))
        self.assertEquals("FeO", accepted[0].__str__())

    def test_NaBrCl(self):
        self.icf.set_nominal_composition(CompositionEntry(
            composition="Na2.1ClBr"))
        self.icf.set_maximum_distance(0.1)
        self.icf.set_max_formula_unit_size(5)

        # Make sure it finds only one (Na2ClBr).
        accepted = self.icf.find_all_compounds()
        self.assertEquals(1, len(accepted))

    def test_Ba2As2S(self):
        self.icf.set_nominal_composition(CompositionEntry(
            composition="Ba2As2S"))
        self.icf.set_maximum_distance(0.35)
        self.icf.set_max_formula_unit_size(7)

        # Make sure it finds Ba4As2S.
        accepted = self.icf.find_all_compounds()
        self.assertTrue(CompositionEntry(composition="Ba4As2S") in accepted)


import unittest
import math
from ....vassal.analysis.APRDFAnalysis import APRDFAnalysis
import numpy.testing as np_tst
from ....vassal.data.Atom import Atom
from ....vassal.data.Cell import Cell

class testAPRDFAnalysis(unittest.TestCase):
    def setUp(self):
        # Make the analysis tool.
        self.tool = APRDFAnalysis()

    def tearDown(self):
        self.tool = None

    def test_eval_distance(self):
        # Define the settings.
        self.tool.set_n_windows(3)
        self.tool.set_cut_off_distance(3.0)

        # Test result.
        np_tst.assert_array_almost_equal([1, 2, 3],
                                         self.tool.get_evaluation_distances())

    def test_APRDF(self):
        # Create a B2 structure with lattice parameter of 1.
        structure = Cell()
        structure.add_atom(Atom([0, 0, 0], 0))
        structure.add_atom(Atom([0.5, 0.5, 0.5], 1))

        self.tool.set_n_windows(2)
        self.tool.set_cut_off_distance(1.0)
        self.tool.set_smoothing_factor(100)

        self.tool.analyze_structure(structure)

        # Trivial: Properties == 0.
        ap_rdf = self.tool.compute_APRDF([0, 0])
        np_tst.assert_array_almost_equal([0.5, 1],
                                         self.tool.get_evaluation_distances())
        np_tst.assert_array_almost_equal([0, 0], ap_rdf)

        # Actual case.
        ap_rdf = self.tool.compute_APRDF([1, 2])

        # Assemble known contributors.
        # [0] -> Number of neighbors * P_i * P_j
        # [1] -> Bond distance
        contributors = []
        contributors.append([2 * 8 * 2 * 1, math.sqrt(3) / 2]) # A-B 1st NN.
        contributors.append([6 * 1 * 1, 1])  # A-A 2nd NN.
        contributors.append([6 * 2 * 2, 1])  # B-B 2nd NN.
        contributors.append([8 * 1 * 1, math.sqrt(3)])  # A-A 3rd NN.
        contributors.append([8 * 2 * 2, math.sqrt(3)])  # B-B 3rd NN.

        eval_dist = [0.5, 1]
        expected_ap_rdf = [sum([c[0] * math.exp(-100 * (c[1] - r)
                        ** 2) for c in contributors]) / 2 for r in eval_dist]
        np_tst.assert_array_almost_equal(expected_ap_rdf, ap_rdf, decimal=3)

    def test_unit_cell_choice(self):
        # Create a B2 structure with lattice parameter of 1.
        structure = Cell()
        structure.add_atom(Atom([0, 0, 0], 0))
        structure.add_atom(Atom([0.5, 0.5, 0.5], 1))

        # Create a 2x1x1 supercell.
        supercell = Cell()
        supercell.set_basis(lengths=[2, 1, 1], angles=[90, 90, 90])
        supercell.add_atom(Atom([0, 0, 0], 0))
        supercell.add_atom(Atom([0.5, 0, 0], 0))
        supercell.add_atom(Atom([0.25, 0.5, 0.5], 1))
        supercell.add_atom(Atom([0.75, 0.5, 0.5], 1))

        self.tool.set_cut_off_distance(3.0)
        self.tool.set_n_windows(10)
        self.tool.set_smoothing_factor(4)


        # Compute the primitive cell AP-RDF.
        self.tool.analyze_structure(structure)
        p_ap_rdf = self.tool.compute_APRDF([1, 2])

        # Compute the supercell AP-RDF.
        self.tool.analyze_structure(supercell)
        sc_ap_rdf = self.tool.compute_APRDF([1, 2])

        # Compare results.
        np_tst.assert_array_almost_equal(p_ap_rdf, sc_ap_rdf)


import inspect

def value(string):
    if string in (None ,False, True):
        return True
    else:
        return False


# example
import pandas as pd
a = inspect.getargspec(pd.read_excel).args
d = inspect.getargspec(pd.read_excel).defaults
d = ['* required'] + list(d)
assert len(a)==len(d)
for i in range(len(a)):
    v = d[i]
    if value(v):
        print "%s = Parameter('%s', %s)" % (a[i], a[i], str(d[i]))
    else:
        print "%s = Parameter('%s', '%s')" % (a[i], a[i], str(d[i]))




import unittest
from math import sqrt
from .....data.materials.CompositionEntry import CompositionEntry
from .....data.utilities.filters.CompositionDistanceFilter import \
    CompositionDistanceFilter

class testCompositionDistanceFilter(unittest.TestCase):
    def test_label(self):
        # Make sample list of CompositionEntry's.
        entries =[CompositionEntry(composition="Fe50Al50"), CompositionEntry(
            composition="Fe55Al45"), CompositionEntry(composition="Fe25Al75")]

        # Make filter.
        cdf = CompositionDistanceFilter()
        cdf.set_target_composition(CompositionEntry(composition="Fe49Al51"))
        cdf.set_distance_threshold(1.1)

        # Test #1.
        res = cdf.label(entries)
        self.assertTrue(res[0])
        self.assertFalse(res[1])
        self.assertFalse(res[2])

        # Test #2.
        cdf.set_distance_threshold(7)
        res = cdf.label(entries)
        self.assertTrue(res[0])
        self.assertTrue(res[1])
        self.assertFalse(res[2])

    def test_distances(self):
        # Test binary distance.
        self.assertAlmostEquals(0.0,
                                CompositionDistanceFilter.compute_distance(
                                    CompositionEntry(composition="Fe"),
                                    CompositionEntry(composition="Fe"),
                                    p=-1), delta=1e-6)

        self.assertAlmostEquals(0.0,
                                CompositionDistanceFilter.compute_distance(
                                    CompositionEntry(composition="Fe"),
                                    CompositionEntry(composition="Fe"),
                                    p=0), delta=1e-6)

        self.assertAlmostEquals(0.0,
                                CompositionDistanceFilter.compute_distance(
                                    CompositionEntry(composition="Fe"),
                                    CompositionEntry(composition="Fe"),
                                    p=2), delta=1e-6)

        self.assertAlmostEquals(0.5,
                                CompositionDistanceFilter.compute_distance(
                                    CompositionEntry(composition="FeO"),
                                    CompositionEntry(composition="Fe"),
                                    p=-1), delta=1e-6)

        self.assertAlmostEquals(2.0,
                                CompositionDistanceFilter.compute_distance(
                                    CompositionEntry(composition="FeO)"),
                                    CompositionEntry(composition="Fe"),
                                    p=0), delta=1e-6)

        self.assertAlmostEquals(sqrt(0.5),
                                CompositionDistanceFilter.compute_distance(
                                    CompositionEntry(composition="FeO"),
                                    CompositionEntry(composition="Fe"),
                                    p=2), delta=1e-6)


# __name__ = "cheml"
__version__ = "0.4.3"
__author__ = ["Mojtaba Haghighatlari (mojtabah@buffalo.edu)", "Johannes Hachmann (hachmann@buffalo.edu)"]

from .wrappers.engine import run as wrapperRUN
# from cheml.notebooks import wrapperGUI

import sys
sys.dont_write_bytecode = True



import types
import pandas as pd
import numpy as np
from ....data.materials.CrystalStructureEntry import CrystalStructureEntry

class EffectiveCoordinationNumberAttributeGenerator:
    """Compute attributes based on the effective coordination number.

    Notes
    -----
    The effective coordination number can be thought of as a face-size-weighted
    coordination number. It is computed by the formula

    .. math:: N_{eff} = \displaystyle\frac{1}{\sum[(\frac{f_i}{SA})^2]}

    where :math: `f_i` is the area of face :math: `i` and :math: `SA` is the
    surface area of the entire cell.

    The effective coordination number has major benefit: stability against the
    additional of a very small face. Small perturbations in atomic positions
    can break symmetry in a crystal, and lead to the introduction of small
    faces. The conventional coordination number treats all faces equally,
    so the coordination number changes even when one of these small faces is
    added.

    One approach in the literature is to first apply a screen on small
    faces (e.g., remove any smaller than 1% of the total face area),
    which still runs into problems with discontinuity for larger displacements.

    Our approach is differentiable with respect to the additional of a small
    face (ask Logan if you want the math), and also captures another
    interesting effect small coordination numbers for Voronoi cells with a
    dispersity in face sizes. For example, BCC has 14 faces on its voronoi
    cell. 8 large faces, and 6 small ones. Our effective face size identifies a
    face size of closer to 8, the commonly-accepted value of the BCC
    coordination number, than 14 reported by the conventional measure.
    Additional, for systems with equal-sized faces (e.g., FCC), this measure
    agrees exactly with conventional reports.

    """

    def mean_abs_dev(self, data):
        """Function to compute the mean absolute deviation of an array-like
        collection of numbers.

        Parameters
        ----------
        data : array-like
            A NumPy array of float values.

        Returns
        -------
        output : float
            The mean absolute deviation.

        """

        n = float(len(data))
        mean = sum(data) / n
        diff = [abs(x - mean) for x in data]
        output = sum(diff) / n
        return output

    def generate_features(self, entries):
        """Function to generate features as mentioned in the class description.

        Parameters
        ----------
        entries : array-like
            Crystal structures for which features are to be generated. A list
            of CrystalStructureEntry's.

        Returns
        ----------
        features : DataFrame
            Features for the given entries. Pandas data frame containing the
            names and values of the descriptors.

        Raises
        ------
        ValueError
            If input is not of type list.
            If items in the list are not CrystalStructureEntry instances.

        """

        # Raise exception if input argument is not of type list of
        # CrystalStructureEntry's.

        if (type(entries) is not types.ListType):
            raise ValueError("Argument should be of type list of "
                             "CrystalStructureEntry's")
        elif (entries and not isinstance(entries[0], CrystalStructureEntry)):
            raise ValueError("Argument should be of type list of "
                             "CrystalStructureEntry's")

        # Initialize lists of feature values and headers for pandas data frame.
        feat_headers = []
        feat_values = []

        feat_headers.append("mean_Coordination")
        feat_headers.append("var_Coordination")
        feat_headers.append("min_Coordination")
        feat_headers.append("max_Coordination")

        l_fh = len(feat_headers)
        for entry in entries:
            temp_list = []
            try:
                output = entry.compute_voronoi_tessellation()
            except Exception:
                tmp_list = [np.nan] * l_fh # If tessellation fails.
                feat_values.append(tmp_list)
                continue
            N_eff = output.get_effective_coordination_numbers()

            mean = np.mean(N_eff)
            absdev = self.mean_abs_dev(data=N_eff)
            minimum = np.min(N_eff)
            maximum = np.max(N_eff)

            temp_list.append(mean)
            temp_list.append(absdev)
            temp_list.append(minimum)
            temp_list.append(maximum)

            feat_values.append(temp_list)

        features = pd.DataFrame(feat_values, columns=feat_headers)
        return features


import unittest
import os
import shutil, tempfile
import numpy as np
import pandas as pd
from cheml.visualization import scatter2D, hist
from cheml.visualization import decorator
from cheml.visualization import SavePlot

# dummy data
x = pd.DataFrame(np.arange(0.0, 1.0, 0.01))
y = pd.DataFrame(np.sin(2*np.pi*x))



class Testvisualization(unittest.TestCase):
    def setUp(self):
        # Create a temporary directory
        self.test_dir = tempfile.mkdtemp()

    def tearDown(self):
        # Remove the directory after the test
        shutil.rmtree(self.test_dir)

    def test_scatter2D(self):
        sc = scatter2D('r', linestyle='--')
        fig = sc.plot(x,y,0,0)
        fig.show()

    def test_hist(self):
        hg = hist(20,'g',{'normed':True})
        fig = hg.plot(y,0)
        fig.show()

    def test_decorator(self):
        hg = hist(20,'g',{'normed':True})
        fig = hg.plot(y,0)
        dec = decorator('histogram',xlabel='sin', ylabel='sin%', xlim=(4,None), ylim=(0,None),
                grid=True, grid_color='g', grid_linestyle=':', grid_linewidth=0.5)
        fig = dec.fit(fig)
        dec.matplotlib_font()
        fig.show()

    def test_SavePlot(self):
        sc = scatter2D('r', linestyle='--')
        fig = sc.plot(x,y,0,0)
        sp = SavePlot('Sin', os.path.join(self.test_dir,'plots'), 'eps', {'facecolor': 'w', 'dpi': 100, 'pad_inches': 0.1, 'bbox_inches': 'tight'})
        sp.save(fig)


if __name__== '__main__':
    unittest.main()




import numpy as np

from keras.models import Sequential
from keras.optimizers import SGD

from json import load
from importlib import import_module

from sklearn.base import BaseEstimator, RegressorMixin


class MLP(object):
    """
    Class associated with Multi-Layer Perceptron (Neural Network)

    Parameters
    ----------
    nhidden : int, optional, default: 1
        The number of hidden layers in the neural network (excluding input and output)

    nneurons: list, optional, default: [100] * nhidden
        The number of nodes in each hidden layer. Must be of same length as nhidden

    activations: list, optional, default: ['sigmoid'] * nhidden
        The activation type for each hidden layer. Must be of same length as nhidden.
        Refer https://keras.io/activations/ for list of valid activations

    nepochs: int, optional, default: 100
        Number of training epochs.

    batch_size: int, optional, default: 100
        Number of training samples in mini-batch

    loss: str, optional, default: 'mean_squared_error'
        Type of loss used to train the neural network.
        Refer https://keras.io/losses/ for list of valid losses

    regression: bool, optional, default: True
        Decides whether we are training for regression or classification task

    nclasses: int, optional, default: None
        Number of classes labels needs to be specified if regression is False

    layer_config_file: str, optional, default: None
        Path to the file that specifies layer configuration
        Refer MLP test to see a sample file

    opt_config_file: str, optional, default: None
        Path to the file that specifies optimizer configuration
        Refer MLP test to see a sample file


    Attributes
    ----------

    Methods
    -------


    """

    def __init__(self, nhidden=1, nneurons=None, activations=None,
                 learning_rate=0.01, lr_decay=0.0,
                 nepochs=100, batch_size=100, loss='mean_squared_error',
                 regression=True, nclasses=None,
                 layer_config_file=None, opt_config_file=None):
        self.model = Sequential()
        if layer_config_file:
            self.layers = self.parse_layer_config(layer_config_file)
        else:
            self.layers = []
            self.nhidden = nhidden
            self.nneurons = nneurons if nneurons else [100] * nhidden
            self.activations = activations if activations else ['sigmoid'] * nhidden
        if opt_config_file:
            self.opt = self.parse_opt_config(opt_config_file)
        else:
            self.opt = SGD(lr=learning_rate, momentum=0.9, decay=lr_decay, nesterov=False)
        self.nepochs = nepochs
        self.batch_size = batch_size
        self.loss = loss
        self.is_regression = regression
        self.nclasses = nclasses

    def fit(self, X, y):
        """Train the MLP for training data X and targets y

        Parameters
        ----------
        X: array_like, shape=[n_samples, n_features]
            Training data

        y: array_like, shape=[n_samples,]
            Training targets

        """
        if len(self.layers) == 0:
            for i in range(self.nhidden):
                self.layers.append(('Dense',
                                    {'units': self.nneurons[i],
                                     'activation': self.activations[i]}))
            if self.is_regression:
                self.layers.append(('Dense',
                                    {'units': 1,
                                     'activation': 'linear'}))
            else:
                self.layers.append(('Dense',
                                    {'units': self.nclasses,
                                     'activation': 'softmax'}))
        layer_name, layer_params = self.layers[0]
        layer_params['input_dim'] = X.shape[-1]
        keras_layer_module = import_module('keras.layers')
        for layer_name, layer_params in self.layers:
            layer = getattr(keras_layer_module, layer_name)
            self.model.add(layer(**layer_params))
        self.model.compile(loss=self.loss, optimizer=self.opt)
        self.batch_size = X.shape[0] if X.shape[0] < self.batch_size else self.batch_size
        self.model.fit(x=X, y=y, epochs=self.nepochs, batch_size=self.batch_size)

    def predict(self, X):
        """
        Return prediction for test data X

        Parameters
        ----------
        X: array_like, shape=[n_samples, n_features]
            Testing data

        Returns
        -------
        float
            Predicted value from model

        """
        return self.model.predict(X).squeeze() if self.is_regression else np.argmax(self.model.predict(X).squeeze())

    def score(self, X, y):
        """
        Predict results for test data X and compare with true targets y. Returns root mean square error if regression,
        accuracy if classification

        Parameters
        ----------
        X: array_like, shape=[n_samples, n_features]
            Test data

        y: array_like, shape=[n_samples,]
            True targets

        Returns
        -------
        float
            root mean square error if regression, accuracy if classification
        """
        prediction = self.model.predict(X).squeeze()
        if self.is_regression:
            return np.mean((prediction - y) ** 2) ** 0.5
        else:
            return np.sum(np.argmax(prediction, axis=1) == y) * 100. / len(y)

    def parse_layer_config(self, layer_config_file):
        """
        Internal method to parse a layer config file

        Parameters
        ----------
        layer_config_file: str
            Filepath that contains the layer configuration file - Refer MLP test to see a sample file
            Refer MLP test to see a sample file and https://keras.io/layers/about-keras-layers/
            for all possible types of layers and corresponding layer parameters

        Returns
        -------
        layers: list
            List of tuples containing layer type and dictionary of layer parameter arguments

        """
        with open(layer_config_file, 'r') as f:
            layers = load(f)
        return layers

    def parse_opt_config(self, opt_config_file):
        """
        Internal method to parse a optimizer config file

        Parameters
        ----------
        opt_config_file: str
            Filepath that contains the optimizer configuration file - Refer MLP test to see a sample file
            Refer MLP test to see a sample file and https://keras.io/optimizers/
            for all possible types of optimizers and corresponding optimizer parameters

        Returns
        -------
        opt: keras.optimizers
            keras optimizer created out of contents of optmizer configuration file

        """
        with open(opt_config_file, 'r') as f:
            opt_name, opt_params = load(f)
        keras_opt_module = import_module('keras.optimizers')
        opt = getattr(keras_opt_module, opt_name)(**opt_params)
        return opt

class MLP_sklearn(BaseEstimator, RegressorMixin):
    """
    A Scikit_learn wrapper around Multi-Layer Perceptron (Neural Network) implemented in keras to be used as
    part of your scikit_learn workflow.

    Parameters
    ----------
    nhidden : int, optional, default: 1
        The number of hidden layers in the neural network (excluding input and output)

    nneurons: list, optional, default: [100] * nhidden
        The number of nodes in each hidden layer. Must be of same length as nhidden

    activations: list, optional, default: ['sigmoid'] * nhidden
        The activation type for each hidden layer. Must be of same length as nhidden.
        Refer https://keras.io/activations/ for list of valid activations

    nepochs: int, optional, default: 100
        Number of training epochs.

    batch_size: int, optional, default: 100
        Number of training samples in mini-batch

    loss: str, optional, default: 'mean_squared_error'
        Type of loss used to train the neural network.
        Refer https://keras.io/losses/ for list of valid losses

    regression: bool, optional, default: True
        Decides whether we are training for regression or classification task

    nclasses: int, optional, default: None
        Number of classes labels needs to be specified if regression is False

    layer_config_file: str, optional, default: None
        Path to the file that specifies layer configuration
        Refer MLP test to see a sample file

    opt_config_file: str, optional, default: None
        Path to the file that specifies optimizer configuration
        Refer MLP test to see a sample file


    Attributes
    ----------

    Methods
    -------


    """
    def __init__(self, nhidden=1, nneurons=None, activations=None,
                 learning_rate=0.01, lr_decay=0.0,
                 nepochs=100, batch_size=100, loss='mean_squared_error',
                 regression=True, nclasses=None,
                 layer_config_file=None, opt_config_file=None):
        self.layer_config_file = layer_config_file
        self.opt_config_file = opt_config_file
        self.layers = []
        self.nhidden = nhidden
        self.nneurons = nneurons
        self.activations = activations
        self.nepochs = nepochs
        self.batch_size = batch_size
        self.loss = loss
        self.regression = regression
        self.nclasses = nclasses
        self.learning_rate = learning_rate
        self.lr_decay = lr_decay
        # model will be defined only after fitting
        self.model = None

    def fit(self, X, y):
        """Train the MLP_sklearn for training data X and targets y

        Parameters
        ----------
        X: array_like, shape=[n_samples, n_features]
            Training data

        y: array_like, shape=[n_samples,]
            Training targets

        """
        self.model = MLP(nhidden=self.nhidden, nneurons=self.nneurons, activations=self.activations,
                 learning_rate=self.learning_rate, lr_decay=self.lr_decay,
                 nepochs=self.nepochs, batch_size=self.batch_size, loss=self.loss,
                 regression=self.regression, nclasses=self.nclasses,
                 layer_config_file=self.layer_config_file, opt_config_file=self.opt_config_file)

        self.model.fit(X,y)


    def predict(self, X):
        """
        Return prediction for test data X

        Parameters
        ----------
        X: array_like, shape=[n_samples, n_features]
            Testing data

        Returns
        -------
        float
            Predicted value from model

        """
        return self.model.predict(X)


    def score(self, X, y, sample_weight=None):
        """
        Predict results for test data X and compare with true targets y. Returns root mean square error if regression,
        accuracy if classification

        Parameters
        ----------
        X: array_like, shape=[n_samples, n_features]
            Test data

        y: array_like, shape=[n_samples,]
            True targets

        Returns
        -------
        float
            root mean square error if regression, accuracy if classification
        """
        prediction = self.predict(X)
        if self.regression:
            return np.mean((prediction - y) ** 2) ** 0.5
        else:
            return np.sum(np.argmax(prediction, axis=1) == y) * 100. / len(y)






"""
The cheml.datasets module includes (please click on links adjacent to function names for more information):
    - load_cep_homo: :func:`~cheml.datasets.load_cep_homo`
    - load_organic_density: :func:`~cheml.datasets.load_organic_density`
    - load_xyz_polarizability: :func:`~cheml.datasets.load_xyz_polarizability`
    - load_comp_energy: :func:`~cheml.datasets.load_comp_energy`
    - load_crystal_structures: :func:`~cheml.datasets.load_crystal_structures`
"""

from .base import load_cep_homo
from .base import load_organic_density
from .base import load_xyz_polarizability
from .base import load_comp_energy
from .base import load_crystal_structures

__all__ = [
    'load_cep_homo',
    'load_organic_density',
    'load_xyz_polarizability',
    'load_comp_energy',
    'load_crystal_structures'
]


# if chemml is not installed yet:
# import sys
# sys.path.insert(0, "/Users/mojtabahaghighatlari/Box Sync/Hachmann_Group/7_MLpackages/chemml")
# sys.path.insert(0, "/projects/academic/hachmann/mojtaba/chemml/")



import itertools


FORK = u'\u251c'
LAST = u'\u2514'
VERTICAL = u'\u2502'
HORIZONTAL = u'\u2500'
NEWLINE = u'\u23ce'


def _format_newlines(prefix, formatted_node):
    """
    Convert newlines into U+23EC characters, followed by an actual newline and
    then a tree prefix so as to position the remaining text under the previous
    line.
    """
    replacement = u''.join([
        NEWLINE,
        u'\n',
        prefix])
    return formatted_node.replace(u'\n', replacement)


def _format_tree(node, format_node, get_children, prefix=u''):
    children = list(get_children(node))
    next_prefix = u''.join([prefix, VERTICAL, u'   '])
    for child in children[:-1]:
        yield u''.join([prefix,
                        FORK,
                        HORIZONTAL,
                        HORIZONTAL,
                        u' ',
                        _format_newlines(next_prefix,
                                         format_node(child))])
        for result in _format_tree(child,
                                   format_node,
                                   get_children,
                                   next_prefix):
            yield result
    if children:
        last_prefix = u''.join([prefix, u'    '])
        yield u''.join([prefix,
                        LAST,
                        HORIZONTAL,
                        HORIZONTAL,
                        u' ',
                        _format_newlines(last_prefix,
                                         format_node(children[-1]))])
        for result in _format_tree(children[-1],
                                   format_node,
                                   get_children,
                                   last_prefix):
            yield result


def format_tree(node, format_node, get_children):
    lines = itertools.chain(
        [format_node(node)],
        _format_tree(node, format_node, get_children),
        [u''],
    )
    return u'\n'.join(lines)


def print_tree(*args, **kwargs):
    print(format_tree(*args, **kwargs))


def test():
    from operator import itemgetter
    tree = (
        'foo', [
            ('bar', [
                ('a', []),
                ('b', []),
            ]),
            ('baz', []),
            ('qux', [
                ('c\nd', []),
            ]),
        ],
    )
    print format_tree(
        tree, format_node=itemgetter(0), get_children=itemgetter(1))


import types
import numpy as np
import pandas as pd
from ....data.materials.CrystalStructureEntry import CrystalStructureEntry
from ....models.regression.crystal.CoulombSineMatrixRegression import \
    CoulombSineMatrixRegression

class CoulombMatrixAttributeGenerator:
    """Class to compute attributes using the Coulomb Sine Matrix
    representation. Based on work by Faber et al. [1].

    Attributes
    ----------
    n_eigenvalues : int
        Maximum number of atoms to consider. Defines number of attributes.

    Notes
    -----
    This method works by computing an approximation for the Coulomb matrix
    that considers periodicity. Specifically, we use the Coulomb Sine matrix,
    which is described in detail in the Faber et al.[1]. For molecules,
    the Coulomb matrix is defined as

    .. math:: C_{i,j} &= Z_i^{2.4} & \text{if} i=j\\
                      &= Z_i Z_j / r_ij & \text{if} i != j

    The eigenvalues of this matrix are then used as attributes. In order to
    provided a fixed number of attributes, the first N attributes are defined
    to be the N eigenvalues from the Coulomb matrix. The remaining attributes
    are defined to be zero.

    The Coulomb Matrix attributes are dependant on unit cell choice.
    Please consider transforming your input crystal structures to the primitive
    cell before using these attributes.

    References
    ----------
    .. [1] F. Faber, A. Lindmaa, O. A. von Lilienfeld, and R. Armiento,
    "Crystal structure representations for machine learning models of
    formation energies," International Journal of Quantum Chemistry,
    vol. 115, no. 16, pp. 1094--1101, Apr. 2015.

    """

    def __init__(self):
        """Function to create instance and initialize fields.
        """

        # Maximum number of atoms to consider. Defines number of attributes.
        self.n_eigenvalues = 30

    def set_n_eigenvalues(self, x):
        """Function to set the number of eigenvalues used in representation.

        Parameters
        ----------
        x : int
            Desired number.

        """

        self.n_eigenvalues = x

    def generate_features(self, entries):
        """Function to generate features as mentioned in the class description.

        Parameters
        ----------
        entries : array-like
            Crystal structures for which features are to be generated. A list
            of CrystalStructureEntry's.

        Returns
        ----------
        features : DataFrame
            Features for the given entries. Pandas data frame containing the
            names and values of the descriptors.

        Raises
        ------
        ValueError
            If input is not of type list.
            If items in the list are not CrystalStructureEntry instances.

        """

        # Initialize list of feature values for pandas data frame.
        feat_values = []

        # Raise exception if input argument is not of type list of
        # CrystalStructureEntry's.
        if (type(entries) is not types.ListType):
            raise ValueError("Argument should be of type list of "
                             "CrystalStructureEntry's")
        elif (entries and not isinstance(entries[0], CrystalStructureEntry)):
            raise ValueError("Argument should be of type list of "
                             "CrystalStructureEntry's")

        # Insert header names here.
        feat_headers = ["CoulombMatrix_Eigenvalue" + str(i) for i in range(
            self.n_eigenvalues)]

        # Create tool to compute eigenvalues.
        tool = CoulombSineMatrixRegression()

        # Generate features.
        for entry in entries:
            eigenvalues = tool.compute_representation(entry.get_structure())
            tmp_array = np.zeros(self.n_eigenvalues, dtype=float)
            if len(eigenvalues) < self.n_eigenvalues:
                tmp_array[:len(eigenvalues)] = eigenvalues[:]
            else:
                tmp_array[:] = eigenvalues[:self.n_eigenvalues]
            feat_values.append(tmp_array)

        features = pd.DataFrame(feat_values, columns=feat_headers)
        return features


import unittest
import os
import numpy as np
import time
from .....vassal.analysis.voronoi.VoronoiTessellationCalculator import \
    VoronoiTessellationCalculator
from .....vassal.data.Atom import Atom
from .....vassal.data.Cell import Cell
import numpy.testing as np_tst
from .....vassal.io.VASP5IO import VASP5IO

class testVoronoiTessellationCalculator(unittest.TestCase):
    this_file_path = os.path.dirname(__file__)
    abs_path = os.path.join(this_file_path, "../../../test-files/")

    def test_simple_cubic(self):
        # Create the simulation cell.
        structure = Cell()
        atom = Atom([0, 0, 0], 0)
        structure.add_atom(atom)

        # Run tessellation.
        result = VoronoiTessellationCalculator.compute(structure,
                                                       radical=False)

        # Test results.
        self.assertEquals(structure.n_atoms(), len(result))
        self.assertEquals(6, len(result[0].get_faces()))
        self.assertAlmostEquals(structure.volume(), result[0].get_volume(),
                                delta=1e-6)
        poly_index = result[0].get_polyhedron_shape()
        self.assertEquals(6, poly_index[4])
        poly_index = result[0].get_coordination_shell_shape(result)
        self.assertEquals(6, poly_index[0])

        # Test out the nearest neighbor shells.
        # 1st NN shell.
        nns = result[0].get_neighbor_shell(result, 1)
        self.assertEquals(6, len(nns))

        # 2nd NN shell.
        nns = result[0].get_neighbor_shell(result, 2)
        self.assertEquals(18, len(nns))

        # 3rd - 5th NN shell.
        for s in range(3, 6):
            nns = result[0].get_neighbor_shell(result, s)
            for image in nns:
                cell = image.get_supercell()
                self.assertAlmostEquals(s, sum([abs(cell[i]) for i in range(
                    3)]), delta=1e-6)

        # Test path NNs.
        # 0th NN.
        paths = result[0].get_neighbors_by_walks(result, 0)
        self.assertEquals(1, len(paths))
        total_weight = sum(paths.values())
        self.assertAlmostEquals(1.0, total_weight, delta=1e-6)

        # 1st NN.
        paths = result[0].get_neighbors_by_walks(result, 1)
        self.assertEquals(6, len(paths))
        total_weight = sum(paths.values())
        np_tst.assert_array_almost_equal([1/6.0]*6, paths.values())
        self.assertAlmostEquals(1.0, total_weight, delta=1e-6)

        # 2nd NN.
        paths = result[0].get_neighbors_by_walks(result, 2)
        self.assertEquals(18, len(paths))
        total_weight = sum(paths.values())
        for k,v in paths.iteritems():
            if 2 in k.get_supercell() or -2 in k.get_supercell():
                self.assertAlmostEquals(1 / 30.0, v, delta=1e-6)
            else:
                self.assertAlmostEquals(2 / 30.0, v, delta=1e-6)
        self.assertAlmostEquals(1.0, total_weight, delta=1e-6)

    def test_BCC(self):
        # Create the simulation cell.
        structure = Cell()
        atom = Atom([0, 0, 0], 0)
        structure.add_atom(atom)
        atom = Atom([0.5, 0.5, 0.5], 0)
        structure.add_atom(atom)

        # Run tessellation.
        result = VoronoiTessellationCalculator.compute(structure,
                                                       radical=False)

        # Test results.
        self.assertEquals(structure.n_atoms(), len(result))
        for cell in result:
            self.assertTrue(cell.geometry_is_valid())
            self.assertEquals(14, len(cell.get_faces()))
            self.assertAlmostEquals(0.5, cell.get_volume(), delta=1e-6)
            poly_index = cell.get_polyhedron_shape()
            self.assertEquals(8, poly_index[6])
            poly_index = result[0].get_coordination_shell_shape(result)
            self.assertEquals(6, poly_index[4])
            self.assertEquals(8, poly_index[6])

    def test_FCC(self):
        # Create the simulation cell.
        structure = Cell()
        structure.add_atom(Atom([0, 0, 0], 0))
        structure.add_atom(Atom([0.5, 0.5, 0], 0))
        structure.add_atom(Atom([0.5, 0, 0.5], 0))
        structure.add_atom(Atom([0, 0.5, 0.5], 0))

        # Run tessellation.
        result = VoronoiTessellationCalculator.compute(structure,
                                                       radical=False)

        # Test results.
        self.assertEquals(structure.n_atoms(), len(result))
        for cell in result:
            self.assertTrue(cell.geometry_is_valid())
            self.assertEquals(12, len(cell.get_faces()))
            self.assertAlmostEquals(0.25, cell.get_volume(), delta=1e-6)
            poly_index = cell.get_polyhedron_shape()
            self.assertEquals(12, poly_index[4])
            poly_index = result[0].get_coordination_shell_shape(result)
            self.assertEquals(12, poly_index[4])

    def test_FCC_primitive(self):
        # Create the simulation cell.
        structure = Cell()
        structure.set_basis(lengths=[0.70710678118655, 0.70710678118655,
                                     1.0], angles=[45, 90, 60])
        structure.add_atom(Atom([0, 0, 0], 0))

        # Run tessellation.
        result = VoronoiTessellationCalculator.compute(structure,
                                                       radical=False)

        # Test results.
        self.assertEquals(structure.n_atoms(), len(result))
        self.assertTrue(result[0].geometry_is_valid())
        self.assertEquals(12, len(result[0].get_faces()))
        poly_index = result[0].get_polyhedron_shape()
        self.assertEquals(12, poly_index[4])
        poly_index = result[0].get_coordination_shell_shape(result)
        self.assertEquals(12, poly_index[4])

    def test_Ta(self):
        structure = VASP5IO.parse_file(file_name=self.abs_path +
                                                 "393-Ta1.vasp")
        # Run tessellation.
        result = VoronoiTessellationCalculator.compute(structure,
                                                       radical=False)

        # Test results.
        total_vol = 0.0
        for cell in result:
            total_vol += cell.get_volume()
            self.assertTrue(cell.geometry_is_valid())
        self.assertAlmostEquals(structure.volume(), total_vol,
                                delta=total_vol * 0.05)

    def test_Zr(self):
        structure = VASP5IO.parse_file(file_name=self.abs_path +
                                                 "1214-Zr1.vasp")
        # Run tessellation.
        result = VoronoiTessellationCalculator.compute(structure,
                                                       radical=False)

        # Test results.
        total_vol = 0.0
        for cell in result:
            total_vol += cell.get_volume()
            self.assertTrue(cell.geometry_is_valid())
        self.assertAlmostEquals(structure.volume(), total_vol,
                                    delta=total_vol * 0.01)

    def test_C(self):
        structure = VASP5IO.parse_file(file_name=self.abs_path +
                                                 "1004-C1.vasp")
        # Run tessellation.
        result = VoronoiTessellationCalculator.compute(structure,
                                                       radical=False)

        # Test results.
        total_vol = 0.0
        for cell in result:
            total_vol += cell.get_volume()
            self.assertTrue(cell.geometry_is_valid())
        self.assertAlmostEquals(structure.volume(), total_vol,
                                    delta=total_vol * 0.01)

    def test_C2(self):
        structure = VASP5IO.parse_file(file_name=self.abs_path + "846-C1.vasp")
        # Run tessellation.
        result = VoronoiTessellationCalculator.compute(structure,
                                                       radical=False)

        # Test results.
        total_vol = 0.0
        for cell in result:
            total_vol += cell.get_volume()
            self.assertTrue(cell.geometry_is_valid())
        vol_error = (total_vol - structure.volume()) / structure.volume()
        self.assertAlmostEquals(0.0, vol_error, delta=1e-2)

    def test_Ho(self):
        structure = VASP5IO.parse_file(file_name=self.abs_path +
                                                 "592-Ho1.vasp")
        # Run tessellation.
        result = VoronoiTessellationCalculator.compute(structure,
                                                       radical=False)

        # Test results.
        total_vol = 0.0
        for cell in result:
            total_vol += cell.get_volume()
            self.assertTrue(cell.geometry_is_valid())
        vol_error = (total_vol - structure.volume()) / structure.volume()
        self.assertAlmostEquals(0.0, vol_error, delta=1e-2)

    def test_Si(self):
        structure = VASP5IO.parse_file(file_name=self.abs_path +
                                                 "399-Si1.vasp")
        # Run tessellation.
        result = VoronoiTessellationCalculator.compute(structure,
                                                       radical=False)

        # Test results.
        total_vol = 0.0
        for cell in result:
            total_vol += cell.get_volume()
            self.assertTrue(cell.geometry_is_valid())
        vol_error = (total_vol - structure.volume()) / structure.volume()
        self.assertAlmostEquals(0.0, vol_error, delta=1e-2)

    def test_Li(self):
        structure = VASP5IO.parse_file(file_name=self.abs_path +
                                                 "478-Li1.vasp")
        # Run tessellation.
        result = VoronoiTessellationCalculator.compute(structure,
                                                       radical=False)

        # Test results.
        total_vol = 0.0
        for cell in result:
            total_vol += cell.get_volume()
            self.assertTrue(cell.geometry_is_valid())
        vol_error = (total_vol - structure.volume()) / structure.volume()
        self.assertAlmostEquals(0.0, vol_error, delta=1e-2)

    def test_B(self):
        structure = VASP5IO.parse_file(file_name=self.abs_path + "673-B1.vasp")
        # Run tessellation.
        result = VoronoiTessellationCalculator.compute(structure,
                                                       radical=False)

        # Test results.
        total_vol = 0.0
        for cell in result:
            total_vol += cell.get_volume()
            self.assertTrue(cell.geometry_is_valid())
        self.assertAlmostEquals(structure.volume(), total_vol,
                                    delta=total_vol * 0.01)

    def test_ICSD_examples(self):
        example = ["3315-Ge2Os2Th1", "1001-N1Y1", "11375-C2N1",
            "12012-Ge2Ru2Tb1", "3778-Sr1Zn2", "4746-Cd1Cu4Er1"]
        for e in example:
            structure = VASP5IO.parse_file(file_name=self.abs_path + e +
                                                     ".vasp")
            # Run tessellation.
            result = VoronoiTessellationCalculator.compute(structure,
                                                           radical=False)

            # Test results.
            total_vol = 0.0
            for cell in result:
                total_vol += cell.get_volume()
                self.assertTrue(cell.geometry_is_valid())
            vol_error = (total_vol - structure.volume()) / structure.volume()
            self.assertAlmostEquals(0.0, vol_error, delta=1e-2)

    def test_Hg2K(self):
        structure = VASP5IO.parse_file(file_name=self.abs_path +
                                                 "7823-Hg2K1.vasp")
        # Run tessellation.
        result = VoronoiTessellationCalculator.compute(structure,
                                                       radical=False)

        # Test results.
        total_vol = 0.0
        for cell in result:
            total_vol += cell.get_volume()
            self.assertTrue(cell.geometry_is_valid())
        vol_error = (total_vol - structure.volume()) / structure.volume()
        self.assertAlmostEquals(0.0, vol_error, delta=1e-2)

    def test_Ag2Pr1Si2(self):
        structure = VASP5IO.parse_file(file_name=self.abs_path +
                                                 "8379-Ag2Pr1Si2.vasp")
        # Run tessellation.
        result = VoronoiTessellationCalculator.compute(structure,
                                                       radical=False)

        # Test results.
        total_vol = 0.0
        for cell in result:
            total_vol += cell.get_volume()
            self.assertTrue(cell.geometry_is_valid())
        vol_error = (total_vol - structure.volume()) / structure.volume()
        self.assertAlmostEquals(0.0, vol_error, delta=1e-2)

    def test_Sc(self):
        structure = VASP5IO.parse_file(file_name=self.abs_path +
                                                 "1565-Sc1.vasp")
        # Run tessellation.
        result = VoronoiTessellationCalculator.compute(structure,
                                                       radical=False)

        # Test results.
        total_vol = 0.0
        for cell in result:
            total_vol += cell.get_volume()
            self.assertTrue(cell.geometry_is_valid())
        vol_error = (total_vol - structure.volume()) / structure.volume()
        self.assertAlmostEquals(0.0, vol_error, delta=1e-2)

    def test_big(self):
        # Number of atoms in each direction.
        n_atom = 10
        structure = Cell()
        structure.set_basis(lengths=[2 * n_atom, 2 * n_atom, 2 * n_atom],
                            angles=[90, 90, 90])

        # Add a bunch of atoms.
        step_size = 1.0 / n_atom
        for x in range(n_atom):
            for y in range(n_atom):
                for z in range(n_atom):
                    new_pos = np.array([x, y, z], dtype=float) + \
                              np.random.random(3) / n_atom
                    new_pos *= step_size
                    structure.add_atom(Atom(new_pos, 0))

        # Compute the cells.
        cells = VoronoiTessellationCalculator.compute(structure, radical=True)
        total_vol = 0.0
        for cell in cells:
            total_vol += cell.get_volume()
            self.assertTrue(cell.geometry_is_valid())

        vol_error = (total_vol - structure.volume()) / structure.volume()
        self.assertAlmostEquals(0.0, vol_error, delta=1e-2)

    def test_random_packing(self):
        # Number of atoms in each direction.
        n_atom = 10
        structure = Cell()
        structure.set_basis(lengths=[2 * n_atom, 2 * n_atom, 2 * n_atom],
                            angles=[90, 90, 90])

        # Add a bunch of atoms.
        for x in range(n_atom):
            for y in range(n_atom):
                for z in range(n_atom):
                    structure.add_atom(Atom(np.random.random(3), 0))

        # Compute the cells.
        cells = VoronoiTessellationCalculator.compute(structure, radical=True)
        total_vol = 0.0
        for cell in cells:
            total_vol += cell.get_volume()
            self.assertTrue(cell.geometry_is_valid())

        vol_error = (total_vol - structure.volume()) / structure.volume()
        self.assertAlmostEquals(0.0, vol_error, delta=1e-2)

    def test_scaling(self):
        print "size\ttime\tvol_error"
        n_atom_list = range(1, 9) + [10, 20]
        # Number of atoms in each direction.
        for n_atom in n_atom_list:
            structure = Cell()
            structure.set_basis(lengths=[2 * n_atom, 2 * n_atom, 2 * n_atom],
                                angles=[90, 90, 90])
            # Add a bunch of atoms.
            step_size = 1.0 / n_atom
            for x in range(n_atom):
                for y in range(n_atom):
                    for z in range(n_atom):
                        new_pos = np.array([x, y, z], dtype=float) + \
                                  np.random.random(3) / n_atom
                        new_pos *= step_size
                        structure.add_atom(Atom(new_pos, 0))

            # Compute the cells.
            start = time.time()
            cells = VoronoiTessellationCalculator.compute(structure,
                                                          radical=True)
            run_time = time.time() - start
            total_vol = 0.0
            for cell in cells:
                total_vol += cell.get_volume()
                self.assertTrue(cell.geometry_is_valid())

            vol_error = (total_vol - structure.volume()) / structure.volume()
            self.assertAlmostEquals(0.0, vol_error, delta=1e-2)
            print "{:d}\t{:.3f}\t{:.2f}".format(structure.n_atoms(),
                    run_time/1000.0, vol_error * 100)

    def test_layered_compound(self):
        structure = VASP5IO.parse_file(file_name=self.abs_path +
                                                 "16234-O2Si1.vasp")
        # Run tessellation.
        result = VoronoiTessellationCalculator.compute(structure,
                                                       radical=False)

        # Test results.
        total_vol = 0.0
        for cell in result:
            total_vol += cell.get_volume()
            self.assertTrue(cell.geometry_is_valid())
        vol_error = (total_vol - structure.volume()) / structure.volume()
        self.assertAlmostEquals(0.0, vol_error, delta=1e-2)


import numpy as np

class Atom:
    """Class that represents a single atom.

    Attributes
    ----------
    position : array-like
        Position in fractional coordinates.
    position_cartesian : array-like
        Position in cartesian coordinates.
    type : int
        Type of atom.
    cell : Cell
        Cell associated with this atom.
    id : int
        ID number of this atom.
    radius : float
        Radius of atom.
    """

    def __init__(self, position, type):
        """Constructor to create a new instance of the object.

        Attributes
        ----------
        position : array-like
            Position in fractional coordinates.
        type : int
            Type of atom.
        """

        # Position in fractional coordinates.
        self.position = None

        # Position in Cartesian coordinates.
        self.position_cartesian = None

        # Type of atom.
        self.type = None

        # Cell associated with this atom.
        self.cell = None

        # ID number of this atom.
        self.id = None

        # Radius of atom.
        self.radius = 1.0

        self.position = np.array(position, dtype=float)
        self.type = type

    def __copy__(self):
        """
        Function to override the copy() method.

        Returns
        -------
        output : Atom
            A new instance with the appropriate properties set.
        """
        x = type(self)(self.__class__)
        x.__dict__.update(self.__dict__)
        x.position = self.position.copy()
        x.position_cartesian = self.position_cartesian.copy()
        return x

    def __eq__(self, other):
        """Function to override the check for equality with another object of
        the same class.

        Parameters
        ----------
        other : Atom
            Other object.

        Returns
        -------
        output : bool
            True if equal, else False.
        """

        if isinstance(other, Atom):
            # Compare type.
            my_name = self.get_type_name()
            your_name = other.get_type_name()

            if not (my_name == "NA" and your_name == "NA"):
                # If type has name, compare names.
                if my_name != your_name:
                    return False
            else:
                # Otherwise, compare index.
                if self.type != other.type:
                    return False

            # Compare positions: Equal if they are closer than 0.001 Cartesian.
            my_pos = self.get_position_cartesian()
            other_pos = other.get_position_cartesian()

            dist = 0.0
            for i in range(3):
                dist += (my_pos[i] - other_pos[i])**2
                if dist > 1e-6:
                    return False

            return True
        return False

    def set_id(self, id):
        """Function to set the ID number of this atom.

        Parameters
        ----------
        id : int
            Desired id.

        """

        self.id = id

    def set_type(self, type):
        """Function to set the type of atom.

        Parameters
        ----------
        type : int
            Index of type.

        """

        self.type = type
        if self.cell is not None:
            while self.cell.n_types() <= self.type:
                self.cell.add_type(None)

    def set_cell(self, cell):
        """Function to define the cell in which this atom is situated.

        Parameters
        ----------
        cell : Cell
            Cell.

        """

        self.cell = cell
        self.update_cartesian_coordinates()

    def update_cartesian_coordinates(self):
        """Function to recompute the cartesian coordinates of this atom.

        """

        self.position_cartesian = self.cell.convert_fractional_to_cartesian(
            self.position)

    def get_cell(self):
        """Function to get the cell that contains this atom.

        Returns
        -------
        output : Cell
            Link to the cell.

        """

        return self.cell

    def get_position(self):
        """Function to get the fractional coordinates of this atom.

        Returns
        -------
        output : array-like
            Fractional coordinates.
        """

        return self.position

    def get_position_cartesian(self):
        """Function to get the cartesian coordinates of this atom.

        Returns
        -------
        output : array-like
            Cartesian coordinates.
        """

        return self.position_cartesian

    def set_radius(self, radius):
        """Function to set the radius of the atom.

        Parameters
        ----------
        radius : float
            Desired radius.

        """

        self.radius = radius

    def get_radius(self):
        """Function to get the radius of this atom.

        Returns
        -------
        output : float
            Radius.
        """

        return self.radius

    def get_id(self):
        """Function to get the ID number of this atom.

        Returns
        -------
        output : int
            ID number.
        """

        return self.id

    def get_type(self):
        """Function to get the type of this atom.

        Returns
        -------
        output : int
            Type of atom.
        """

        return self.type

    def get_type_name(self):
        """Function to get the name of this atom type.

        Returns
        -------
        output : str
            Name of the type of atom.
        """

        return self.cell.get_type_name(self.type)


import pkg_resources
import os
import pandas as pd


def load_cep_homo():
    """Load and return a small sample of HOMO energies of organic photovoltaic candidates from CEP database (regression).
    Clean Energy Project (CEP) database is available at: https://cepdb.molecularspace.org
    The unit of HOMO (highest occupied molecular orbitals) energies is electron Volt (eV).
    The photovaltaic candidates are provided using SMILES representation.

    =================   ==============
    rows                           500
    Columns                          2
    headers             smiles,homo_eV
    molecules rep.              SMILES
    Features                         0
    Returns               2 dataframes
    =================   ==============

    Returns
    -------
    returns two dataframes
        - first dataframe for SMILES representation of molecules, shape: (500,1)
        - second dataframe for HOMO energies of the molecules (eV), shape: (500,1)

    Examples
    --------
    >>> from cheml.datasets import load_cep_homo
    >>> smi, homo  = load_cep_homo()
    >>> print list(smi.columns)
    ['smiles']
    >>> print homo.shape
    (500, 1)
    """
    DATA_PATH = pkg_resources.resource_filename('cheml', os.path.join('datasets','data','cep_homo.csv'))
    df = pd.read_csv(DATA_PATH)

    smi = pd.DataFrame(df['smiles'], columns=['smiles'])
    homo = pd.DataFrame(df['homo_eV'], columns=['homo_eV'])

    return smi, homo


def load_organic_density():
    """Load and return 500 small organic molecules with their density and molecular descriptors.

    =================   ======================
    rows                                   500
    Columns                                202
    last twoo headers     smiles,density_Kg/m3
    molecules rep.                      SMILES
    Features                               200
    Returns                       3 dataframes
    =================   ======================

    Returns
    -------
    returns three dataframes
        - first dataframe for SMILES representation of molecules, shape: (500,1)
        - second dataframe for the density of molecules (Kg/m3), shape: (500,1)
        - third dataframe for the molecular descriptors of molecules, shape: (500,1)

    Examples
    --------
    >>> from cheml.datasets import load_organic_density
    >>> smi, density, features = load_organic_density()
    >>> print list(smi.columns)
    ['smiles']
    >>> print features.shape
    (500, 200)
    """
    DATA_PATH = pkg_resources.resource_filename('cheml', os.path.join('datasets','data','moldescriptor_density_smiles.csv'))
    df = pd.read_csv(DATA_PATH)

    smi = pd.DataFrame(df['smiles'], columns=['smiles'])
    density = pd.DataFrame(df['density_Kg/m3'], columns=['density_Kg/m3'])
    features = df.drop(['smiles', 'density_Kg/m3'],1)

    return smi, density, features


def load_xyz_polarizability():
    """Load and return xyz files and polarizability (Bohr^3).
    The xyz coordinates of small organic molecules are optimized with BP86/def2svp level of theory.
    Polarizability of the molecules are also calcualted in the same level of thoery.

    =================   ======================
    rows                                    50
    Columns                                  1
    header                      polarizability
    molecules rep.                         xyz
    Features                                 0
    Returns             1 dataframe and 1 dict
    =================   ======================

    Returns
    -------
    returns a dictionary and a dataframe
        - dictionary: the xyz coordinates and atomic numbers of each atom of the molecule will be in a numpy array.
        - dataframe: the polarizability of each molecule
    """
    DATA_PATH = pkg_resources.resource_filename('cheml', os.path.join('datasets','data','organic_xyz'))
    from ..initialization import XYZreader
    reader = XYZreader(path_pattern=['[1-9]_opt.xyz', '[1-9][0-9]_opt.xyz'],
                       path_root=DATA_PATH,
                       reader='manual',
                       skip_lines=[2, 0])
    coordinates = reader.read()
    df = pd.read_csv(os.path.join(DATA_PATH,'pol.csv'))
    return coordinates, df

def load_comp_energy():
    """Load and return composition entries and formation energies (eV).
    From Magpie https://bitbucket.org/wolverton/magpie

    =================   ======================
    rows                                   630
    header                    formation_energy
    molecules rep.                 composition
    Features                                 0
    Returns             1 dataframe and 1 list
    =================   ======================

    Returns
    -------
    returns a list and a dataframe
        - list: the list of composition entries from CompositionEntry class.
        - dataframe: the formation energy for each composition.
    """
    DATA_PATH = pkg_resources.resource_filename('cheml', os.path.join('chem','magpie_python','test','test-files','small_set_comp.txt'))
    TARGET_PATH = pkg_resources.resource_filename('cheml', os.path.join('chem','magpie_python','test','test-files','small_set_delta_e.txt'))
    from ..chem.magpie_python import CompositionEntry
    entries = CompositionEntry.import_composition_list(DATA_PATH)
    df = pd.read_csv(TARGET_PATH,header=None)
    df.columns = ['formation_energy']
    return entries, df

def load_crystal_structures():
    """Load and return crystal structure entries.
    From Magpie https://bitbucket.org/wolverton/magpie

    =================   ======================
    length                                  18
    header                    formation_energy
    molecules rep.                 composition
    Features                                 0
    Returns                             1 list
    =================   ======================

    Returns
    -------
    returns a list
        - list: the list of crystal structure entries from CrystalStructureEntry class.
    """
    DATA_PATH = pkg_resources.resource_filename('cheml', os.path.join('chem','magpie_python','test','test-files'))
    from ..chem.magpie_python import CrystalStructureEntry
    entries = CrystalStructureEntry.import_structures_list(DATA_PATH)
    return entries




import os
import sys
from ..database import sklearn_db, cheml_db
from ..database.TSHF import tshf
import copy

def txt(ntab=1, text1='', ntab12=1, adj12=False, text2=None, ntab23=1, adj23=False, text3=None,
        ntab34=1, adj34=False, text4=None, ntab45=1, adj45=False, text5=None):
    tab = '    '
    line = tab*ntab + text1
    if text2 is not None:
        if adj12:
            line += (ntab12*4 - len(text1))*' '
        else:
            line += tab*ntab12
        line += text2
        if text3 is not None:
            if adj23:
                line += (ntab23 * 4 - len(text2)) * ' '
            else:
                line += tab * ntab23
            line += text3
            if text4 is not None:
                if adj34:
                    line += (ntab34 * 4 - len(text3)) * ' '
                else:
                    line += tab * ntab34
                line += text4
                if text5 is not None:
                    if adj45:
                        line += (ntab45 * 4 - len(text4)) * ' '
                    else:
                        line += tab * ntab45
                    line += text5

    line = line.rstrip("\n")
    return line

class bcolors:
    HEADER = '\033[95m'
    OKBLUE = '\033[94m'
    OKGREEN = '\033[92m'
    WARNING = '\033[93m'
    FAIL = '\033[91m'
    ENDC = '\033[0m'
    BOLD = '\033[1m'
    UNDERLINE = '\033[4m'

def transform(blocks):
    """
    goals:
        - collect all sends and receives
        - check send and receive format.
        - make the computational graph
        - find the order of implementation of functions based on sends and receives

    :param cmls:
    :return implementation order:
    """
    cmls = [blocks[key] for key in sorted(blocks.iterkeys())]
    send_all = []
    recv_all = []
    for block in cmls:
        send_all += block['send'].items()
        recv_all += block['recv'].items()

    # make graph
    reformat_send = {k[1]: [v, k[0]] for k, v in send_all}
    CompGraph = tuple([tuple(reformat_send[k[1]] + [v, k[0]]) for k, v in recv_all])

    # find orders
    ids_sent = []
    ImpOrder = []
    inf_checker = 0
    while len(ImpOrder) < len(cmls):
        inf_checker += 1
        for i in range(len(cmls)):
            if i not in ImpOrder:
                ids_recvd = [k[1] for k, v in cmls[i]['recv'].items()]
                if len(ids_recvd) == 0:
                    ids_sent += [k[1] for k, v in cmls[i]['send'].items()]
                    ImpOrder.append(i)
                elif len(set(ids_recvd) - set(ids_sent)) == 0:
                    ids_sent += [k[1] for k, v in cmls[i]['send'].items()]
                    ImpOrder.append(i)
        if inf_checker > len(cmls):
            msg = 'Your design of send and receive tokens makes a loop of interdependencies. You can avoid such loops by designing your workflow hierarchichally'
            raise IOError(msg)
    return tuple(ImpOrder), CompGraph


class ui(object):
    def __init__(self):
        self.tasks, self.combinations = tshf()
        self.blocks = {}
        self.block_id = -1
        self.comp_graph = []
        self.out_dir = "CMLWrapper.out"
        self.home_page()


    def home_page(self):
        os.system('cls' if os.name == 'nt' else 'clear')
        print '\n'
        print txt(1, "home           tasks           subtasks           hosts           functions               parameters           pipelines")
        print txt(1, "  @..............:.................:................:.................:........................:...................:")
        print '\n'
        print bcolors.BOLD + txt(11, "Welcome to the ChemML Wrapper Interface") + '\n' + bcolors.ENDC
        print txt(9, "*** Please enlarge the terminal window for a better view ***") + '\n'
        print txt(2, "enter n to start with a new script")
        print txt(2, "enter e to load an existing script")
        print txt(2, "enter t to start from a template")
        key = False
        while not key:
            k = raw_input(bcolors.BOLD + txt(2, "Choose how to start (n : new  ,  e : existing  ,  t : template): ") + bcolors.ENDC)
            k = k.strip()
            if k=='n':
                ko = raw_input(bcolors.BOLD + txt(2, "enter the name of output directory (or press enter for the current name = '%s'): "%self.out_dir) + bcolors.ENDC)
                ko = ko.strip()
                if ko == '':
                    key = True
                else:
                    self.out_dir = ko
                    key = True
            elif k=='e':
                key = True
            elif k=='t':
                key = True
        self.task_page()


    def task_page(self):
        os.system('cls' if os.name == 'nt' else 'clear')
        print '\n'
        print txt(1, "home           tasks           subtasks           hosts           functions               parameters           pipelines")
        print txt(1, "  :..............@.................:................:.................:........................:...................:")
        print '\n'
        print txt(2, "pick a task ...")+ '\n'
        print txt(3, 'id', 1, True, 'task')
        print txt(3, '--', 1, True, '----')
        for i,task in enumerate(self.tasks):
            print txt(3, str(i), 1, True, task)
        print '\n'
        options = [str(i) for i in range(len(self.tasks))] + ['s','q','b', 'p']
        key = False
        while not key:
            k = raw_input(bcolors.BOLD + txt(1,"enter one of ( id ,  p : pipeline page  ,  b : back  ,  s : save  ,  q : quit ): ") + bcolors.ENDC)
            k = k.strip()
            if k in options:
                key = True
        if k == 'q':
            sys.exit()
        elif k == 's':
            os.system('cls' if os.name == 'nt' else 'clear')
        elif k == 'b':
            self.home_page()
        elif k == 'p':
            self.graph_page()
        elif k in [str(i) for i in range(len(self.tasks))]:
            # os.system('cls' if os.name == 'nt' else 'clear')
            self.task = '%s'%self.tasks[int(k)]
            self.subtask_page()


    def subtask_page(self):
        os.system('cls' if os.name == 'nt' else 'clear')
        print '\n'
        print txt(1, "home           tasks           subtasks           hosts           functions               parameters           pipelines")
        print txt(1, "  :..............:.................@................:.................:........................:...................:")
        print '\n'
        subtasks = [i for i in self.combinations[self.task]]
        print txt(1, "task:", 3, True, self.task ) + '\n'
        print txt(2, "pick a subtask ...")+ '\n'
        print txt(3, 'id', 1, True, 'subtask')
        print txt(3, '--', 1, True, '----')
        for i, subtask in enumerate(subtasks):
            print txt(3, str(i), 1, True, subtask)
        print '\n'
        options = [str(i) for i in range(len(subtasks))] + ['b', 'q']
        key = False
        while not key:
            k = raw_input(bcolors.BOLD + txt(1, "enter one of (id, 'b':back, 'q':quit): ") + bcolors.ENDC)
            k = k.strip()
            if k in options:
                key = True
        if k == 'q':
            sys.exit()
        elif k == 'b':
            self.task_page()
        elif k in [str(i) for i in range(len(subtasks))]:
            # os.system('cls' if os.name == 'nt' else 'clear')
            self.subtask = '%s' % subtasks[int(k)]
            self.host_page()


    def host_page(self):
        os.system('cls' if os.name == 'nt' else 'clear')
        print '\n'
        print txt(1, "home           tasks           subtasks           hosts           functions               parameters           pipelines")
        print txt(1, "  :..............:.................:................@.................:........................:...................:")
        print '\n'
        hosts = [i for i in self.combinations[self.task][self.subtask]]
        print txt(1, "task:", 3, True, self.task )
        print txt(1, "subtask:", 3, True , self.subtask) + '\n'
        print txt(2, "pick a host ...")+ '\n'
        print txt(3, 'id', 1, True, 'host')
        print txt(3, '--', 1, True, '----')
        for i, host in enumerate(hosts):
            print txt(3, str(i), 1, True, host)
        print '\n'
        options = [str(i) for i in range(len(hosts))] + ['b', 'q']
        key = False
        while not key:
            k = raw_input(bcolors.BOLD + txt(1, "enter one of (id, 'b':back, 'q':quit): ") + bcolors.ENDC)
            k = k.strip()
            if k in options:
                key = True
        if k == 'q':
            sys.exit()
        elif k == 'b':
            self.subtask_page()
        elif k in [str(i) for i in range(len(hosts))]:
            # os.system('cls' if os.name == 'nt' else 'clear')
            self.host = '%s' % hosts[int(k)]
            self.function_page()


    def function_page(self):
        os.system('cls' if os.name == 'nt' else 'clear')
        print '\n'
        print txt(1, "home           tasks           subtasks           hosts           functions               parameters           pipelines")
        print txt(1, "  :..............:.................:................:.................@........................:...................:")
        print '\n'
        functions = [i for i in self.combinations[self.task][self.subtask][self.host]]
        print txt(1, "task:", 3, True, '%s' % self.task)
        print txt(1, "subtask:", 3, True , '%s' % self.subtask)
        print txt(1, "host:", 3, True , '%s' % self.host) + '\n'
        print txt(2, "pick a function ...")
        print txt(3, 'id', 1, True, 'function')
        print txt(3, '--', 1, True, '----')
        for i, func in enumerate(functions):
            print txt(3, str(i), 1, True, func)
        print '\n'
        options = [str(i) for i in range(len(functions))] + ['b', 'q']
        key = False
        while not key:
            k = raw_input(bcolors.BOLD + txt(1, "enter one of (id, 'b':back, 'q':quit): ") + bcolors.ENDC)
            k = k.strip()
            if k in options:
                key = True
        if k == 'q':
            sys.exit()
        elif k == 'b':
            self.host_page()
        elif k in [str(i) for i in range(len(functions))]:
            # os.system('cls' if os.name == 'nt' else 'clear')
            self.function = '%s' % functions[int(k)]
            self.db_extract_function()
            self.custom_function_page()


    ##################################


    def db_extract_function(self):
        if self.host == 'sklearn':
            self.metadata = getattr(sklearn_db, self.function)()
        elif self.host == 'cheml':
            self.metadata = getattr(cheml_db, self.function)()
        self.wparams = {i:copy.deepcopy(vars(self.metadata.WParameters)[i]) for i in vars(self.metadata.WParameters).keys() if
                     i not in ('__module__', '__doc__')}
        self.fparams = {i:copy.deepcopy(vars(self.metadata.FParameters)[i]) for i in vars(self.metadata.FParameters).keys() if
                     i not in ('__module__', '__doc__')}
        self.parameters = {}
        self.poptions = []
        for i,j in enumerate(self.wparams):
            self.parameters['%iw'%i] = self.wparams[j]
            self.poptions.append('%iw'%i)
        for i,j in enumerate(self.fparams):
            self.parameters['%if'%i] = self.fparams[j]
            self.poptions.append('%if' % i)


    def custom_function_help(self):
        os.system('cls' if os.name == 'nt' else 'clear')
        print '\n'
        print bcolors.BOLD + txt(1, "parameters page keys") + bcolors.ENDC
        print txt(2, bcolors.BOLD + 'a' + bcolors.ENDC, 1, True, "    add function to the workflow as a node.")
        print txt(2, bcolors.BOLD + 'id' + bcolors.ENDC, 1, True, "   edit a parameter's value. This is the parameter id. Next enter the python format of the new value (e.g. string with quotation mark).")
        print txt(2, bcolors.BOLD + 't' + bcolors.ENDC, 1, True, "    go to the task page and ignore this function.")
        print txt(2, bcolors.BOLD + 'b' + bcolors.ENDC, 1, True, "    back to the previous page (functions page).")
        print txt(2, bcolors.BOLD + 'q' + bcolors.ENDC, 1, True, "    quit the ChemML wrapper's interface without saving.")
        print txt(2, bcolors.BOLD + 'h' + bcolors.ENDC, 1, True, "    help page, this page!")
        print '\n'

        key = False
        while not key:
            k = raw_input(bcolors.BOLD + txt(1, "enter q to quit help page: ") + bcolors.ENDC)
            k = k.strip()
            if k in ['q']:
                self.custom_function_page()


    def custom_function_keys(self):
        key = False
        print txt(2, "** edit values by entering parameter's id and new value separately")
        print txt(2, "*** when you are done with editing enter 'a' to add this function to the workflow") + '\n'
        while not key:
            k = raw_input(bcolors.BOLD + txt(1, "enter one of ( id new_value  ,  a : add  ,  b : back  ,  t : tasks page  ,  h : help  ,  q : quit): ") + bcolors.ENDC)
            k = k.strip()
            if k in ['b', 'q', 'a', 't', 'h']:
                key = True
            elif ' ' in k:
                ks = k.split(' ')
                while '' in ks:
                    ks.remove('')
                if len(ks) == 2 and ks[0].strip() in self.parameters:
                    try:
                        param_value = eval(ks[1])
                        key = True
                    except Exception:
                        print bcolors.FAIL+ txt(1,"wrong format: %s - hint: use quotation mark for string"%ks[1]) +bcolors.ENDC

        if k == 'q':
            sys.exit()
        elif k == 'b':
            self.function_page()
        elif k == 'h':
            self.custom_function_help()
        elif k == 'a':
            self.block_id += 1
            self.blocks[self.block_id] = {'function':self.function,'host':self.host,'subtask':self.subtask,'task':self.task,
                                         'parameters': {self.parameters[z].name:self.parameters[z].default for z in self.parameters},
                                         'send': {}, 'recv': {}
                                          }
            self.graph_page()
        elif k == 't':
            self.task_page()
        elif ks[0] in self.parameters:
            self.parameters[ks[0]].default = param_value
            self.custom_function_page()


    def custom_function_page(self):
        os.system('cls' if os.name == 'nt' else 'clear')
        print '\n'
        print txt(1, "home           tasks           subtasks           hosts           functions               parameters           pipelines")
        print txt(1, "  :..............:.................:................:.................:........................@...................:")
        print '\n'
        print txt(1, "task:", 5, True, '%s' % self.task)
        print txt(1, "subtask:", 5, True , '%s' % self.subtask)
        print txt(1, "host:", 5, True , '%s' % self.host)
        print txt(1, "function:", 5, True , '%s' % self.function) + '\n'
        print txt(1, "requirements:", 5, True, '%s'%str(self.metadata.requirements))
        print txt(1, "documentation:", 5, True, '%s'%self.metadata.documentation) + '\n'
        print txt(2, "* parameters with their default values:") + '\n'
        print txt(3, 'id', 2, True, 'name',5,True,'value',5,True,'options')
        print txt(3, '--', 2, True, '----',5,True,'-----',5,True,'-------')
        for i in self.poptions:
            print txt(3,i,2,True,self.parameters[i].name,5,True,str(self.parameters[i].default),
                      5,True, str(self.parameters[i].options))
        print '\n'
        self.custom_function_keys()


    #################################


    def IO(self, ib):
        host = self.blocks[ib]['host']
        function = self.blocks[ib]['function']
        if host == 'sklearn':
            metadata = getattr(sklearn_db, function)()
        elif host == 'cheml':
            metadata = getattr(cheml_db, function)()
        self.inputs = {i:copy.deepcopy(vars(metadata.Inputs)[i]) for i in vars(metadata.Inputs).keys() if
                       i not in ('__module__','__doc__')}
        self.outputs = {i:copy.deepcopy(vars(metadata.Outputs)[i]) for i in vars(metadata.Outputs).keys() if
                        i not in ('__module__', '__doc__')}


    def graph_help(self):
        os.system('cls' if os.name == 'nt' else 'clear')
        print '\n'
        print bcolors.BOLD + txt(1, "general hints") + bcolors.ENDC
        print txt(2, "The list of nodes might have same functions (with different/same parameter values) for different parts of the workflow.")
        print txt(2, "The id of nodes are unique.")
        print txt(2, "After pipelining the nodes, you can see the tree design of the workflow.")
        print '\n'
        print bcolors.BOLD + txt(1, "pipelines page keys") + bcolors.ENDC
        print txt(2, bcolors.BOLD + 'd' + bcolors.ENDC, 1, True, "    delete a node. Next enter a node id with empty space in between (e.g. d 7)")
        print txt(2, bcolors.BOLD + 'id' + bcolors.ENDC, 1, True, "   initiate a pipeline. This is the sender node id. Next enter a receiver node id (e.g. 3 7)")
        print txt(2, '', 1, True, " Next we show you all the sender's output tokens and the receiver's input tokens. enter one token of each respectively.")
        print txt(2, bcolors.BOLD + 't' + bcolors.ENDC, 1, True, "    go to the task page.")
        print txt(2, bcolors.BOLD + 's' + bcolors.ENDC, 1, True, "    save the current workflow on the disk.")
        print txt(2, bcolors.BOLD + 'q' + bcolors.ENDC, 1, True, "    quit the ChemML wrapper's interface without saving.")
        print txt(2, bcolors.BOLD + 'h' + bcolors.ENDC, 1, True, "    help page, this page!")
        print '\n'

        key = False
        while not key:
            k = raw_input(bcolors.BOLD + txt(1, "enter q to quit page: ") + bcolors.ENDC)
            k = k.strip()
            if k in ['q']:
                self.graph_page()


    def No_short_loops(self):
        return True


    def graph_page_pipeline_keys(self,ks):
        print '\n'
        ibs = int(ks[0])
        ibr = int(ks[1])
        sender_function = "%s outputs" % self.blocks[ibs]['function']
        receiver_function = "%s inputs" % self.blocks[ibr]['function']
        print txt(2, sender_function + " " * 8 + ">>>>" + " " * 8 + receiver_function)
        print txt(2, "-" * len(sender_function) + " " * 20 + "-" * len(receiver_function))

        self.IO(ibs)
        sender_outputs = [self.outputs[i].name for i in self.outputs]
        self.IO(ibr)
        receiver_inputs = [self.inputs[i].name for i in self.inputs]
        outputs = ", ".join(sender_outputs)
        inputs = ", ".join(receiver_inputs)
        print txt(2, outputs + " " * (20 + len(sender_function) - len(outputs)) + inputs) + '\n'
        key = False
        while not key:
            k = raw_input(bcolors.BOLD + txt(2,"enter sender output and receiver input ('token token' or enter c to cancel): ") + bcolors.ENDC)
            k = k.strip()
            if k == 'c':
                self.graph_page()
            elif ' ' in k:
                tokens = k.split(' ')
                while '' in ks:
                    tokens.remove('')
                if tokens[0] in sender_outputs and tokens[1] in receiver_inputs:
                    new_edge = [ibs, tokens[0], ibr, tokens[1]]
                    if new_edge in self.comp_graph:
                        print bcolors.OKGREEN + txt(2, "this edge is already in the workflow tree") + bcolors.ENDC
                    else:
                        self.comp_graph.append(new_edge)
                        if self.No_short_loops():
                            key = True
                        else:
                            self.comp_graph.remove(new_edge)
                            print bcolors.FAIL + txt(2, "Exception (short loop): You arenot allowed to make interdependent nodes") + bcolors.ENDC

                    print self.comp_graph
        self.graph_page()


    def graph_page_keys(self):
        key = False
        print txt(1, "*** keys: ( id id : pipeline  ,  d id : delete  ,  t : task page  ,  s : save,  h : help  ,  q : quit )")
        while not key:
            k = raw_input(bcolors.BOLD + txt(1, "enter one of the above keys: ") + bcolors.ENDC)
            k = k.strip()
            if k in ['t', 'q', 's', 'h']:
                key = True
            elif ' ' in k:
                ks = k.split(' ')
                ibs = [str(ib) for ib in self.blocks]
                while '' in ks:
                    ks.remove('')
                if ks[0]=='d' and len(ks)==2:
                    if ks[1] in ibs:
                        key = True
                elif len(ks) == 2 and ks[0] in ibs and ks[1] in ibs:
                    if ks[0] == ks[1]:
                        print bcolors.FAIL + txt(1, "Exception (short loop): You arenot allowed to send info from one node to itself") + bcolors.ENDC
                        key = True # remove later
                    else:
                        key = True
        if k == 'q':
            sys.exit()
        elif k == 's':
            sys.exit()
        elif k == 't':
            self.task_page()
        elif k == 'h':
            self.graph_help()
        elif ks[0]=='d':
            del self.blocks[int(ks[1])]
            self.graph_page()
        else:
            self.graph_page_pipeline_keys(ks)


    def graph_page(self):
        os.system('cls' if os.name == 'nt' else 'clear')
        print '\n'
        print txt(1, "home           tasks           subtasks           hosts           functions               parameters           pipelines")
        print txt(1, "  :..............:.................:................:.................:........................:...................@")
        print '\n'
        print txt(1, "* list of added nodes:") + '\n'
        print txt(2, 'id', 2, True, 'host', 3, True, 'function', 5, True,'input tokens', 9, True, 'output tokens')
        print txt(2, '--', 2, True, '----', 3, True, '--------', 5, True,'------------', 9, True, '-------------')
        for ib in sorted(self.blocks.iterkeys()):
            self.IO(ib)
            host = self.blocks[ib]['host']
            function = self.blocks[ib]['function']
            inputs = [self.inputs[i].name for i in self.inputs]
            outputs = [self.outputs[i].name for i in self.outputs]
            print txt(2, str(ib), 2, True, host, 3, True, function, 5, True, ", ".join(inputs), 9, True, ", ".join(outputs))
        print '\n'
        print txt(1, "** workflow tree:") + '\n'

        self.graph_page_keys()


